%include "x86inc.asm"
%include "x86util.asm"

SECTION .text

%if ARCH_X86_64 == 0

;-----------------------------------------------------------------------------
; 
; Some function rely on a aligned stack pointer. This means that they assume
; that esp is in xxxxxxxCh (must end with 0Ch). It seems that only GCC is able
; to align the stack in that way. This macro can create a stub which rearranges
; the parameters on the stack, then calls into the function (which requires the
; stack alignment) and on return corrects the stack pointer again and returns to
; the caller.
;
; +-------------------------+   
; | return address          |   <----- esp
; +-------------------------+
; | argument 1              |
; +-------------------------+
; | argument 2              |
; +-------------------------+
; | argument 3              |
; +-------------------------+
;
; esp is 4-byte aligned, i.e. it may end with 00h, 04h, 08h or 0Ch. In the
; latter case, we have nothing to do and directly call into the wrapped 
; function. In the other cases we copy the arguments to an appropriate place:
;
;
;                                   +-------------------------+   
;                                   | return stub             | <----- esp (xxxCh)
; +-------------------------+       +-------------------------+   
; | return address          |       | argument 1              |
; +-------------------------+       +-------------------------+
; | argument 1              |       | argument 2              |  
; +-------------------------+  -->  +-------------------------+
; | argument 2              |       | argument 3              |
; +-------------------------+       +-------------------------+
; | argument 3              |       | original return addr.   |
; +-------------------------+       +-------------------------+
;
; The return stub corrects the stack pointer again to its original 
; location and jumps back to the original return address.
;
; first argument : the name of the function this stub will call
; second argument: the number of arguments (of the original function)
%macro STACKALIGNSTUB 2
cextern %1			; declare the method for which we create a stub as external
; can only use the scratch registers: eax, ecx, edx
mov edx,esp
and edx,0fh
cmp edx,0ch
jne .needToAlign
jmp %1
.needToAlign:
mov edx,esp
and edx,0fffffff0h	; edx now points to where the first argument must be placed
mov ecx,[esp]		; save the return address

%assign i 1
%rep 16				; repeat 16 times at max
%if i > %2
	%exitrep
%endif
mov eax,[esp+4*i]
mov [edx+4*i-4],eax
%assign i i+1
%endrep

mov [esp+4*%2],ecx	; store the return address

and esp,0fh
cmp esp,0
jne .l1
lea eax,[.retstub0]
jmp .l3
.l1:
cmp esp,04h
jne .l2
lea eax,[.retstub4]
jmp .l3
.l2:	; eax must now be 08h
lea eax,[.retstub8]
.l3:
mov esp,edx
sub esp,4
mov [esp],eax
jmp %1

.retstub0:
add esp,4
jmp [esp-4+4*%2]

.retstub4:
add esp,8
jmp [esp-4+4*%2]

.retstub8:
add esp,12
jmp [esp-4+4*%2]

%endmacro


;-----------------------------------------------------------------------------


cglobal alignstubbed_intra_satd_x9_4x4_avx
STACKALIGNSTUB intra_satd_x9_4x4_avx,3

cglobal alignstubbed_intra_sad_x9_4x4_avx
STACKALIGNSTUB intra_sad_x9_4x4_avx,3

cglobal alignstubbed_intra_sad_x9_8x8_avx
STACKALIGNSTUB intra_sad_x9_8x8_avx,5

cglobal alignstubbed_pixel_hadamard_ac_16x16_sse2
STACKALIGNSTUB pixel_hadamard_ac_16x16_sse2,2

cglobal alignstubbed_pixel_hadamard_ac_16x8_sse2
STACKALIGNSTUB pixel_hadamard_ac_16x8_sse2,2

cglobal alignstubbed_pixel_hadamard_ac_8x16_sse2
STACKALIGNSTUB pixel_hadamard_ac_8x16_sse2,2

cglobal alignstubbed_pixel_hadamard_ac_8x8_sse2
STACKALIGNSTUB pixel_hadamard_ac_8x8_sse2,2

cglobal alignstubbed_pixel_hadamard_ac_16x16_ssse3
STACKALIGNSTUB pixel_hadamard_ac_16x16_ssse3,2

cglobal alignstubbed_pixel_hadamard_ac_16x8_ssse3
STACKALIGNSTUB pixel_hadamard_ac_16x8_ssse3,2

cglobal alignstubbed_pixel_hadamard_ac_8x16_ssse3
STACKALIGNSTUB pixel_hadamard_ac_8x16_ssse3,2

cglobal alignstubbed_pixel_hadamard_ac_8x8_ssse3
STACKALIGNSTUB pixel_hadamard_ac_8x8_ssse3,2

cglobal alignstubbed_pixel_hadamard_ac_16x16_sse4
STACKALIGNSTUB pixel_hadamard_ac_16x16_sse4,2

cglobal alignstubbed_pixel_hadamard_ac_16x8_sse4
STACKALIGNSTUB pixel_hadamard_ac_16x8_sse4,2

cglobal alignstubbed_pixel_hadamard_ac_8x16_sse4
STACKALIGNSTUB pixel_hadamard_ac_8x16_sse4,2

cglobal alignstubbed_pixel_hadamard_ac_8x8_sse4
STACKALIGNSTUB pixel_hadamard_ac_8x8_sse4,2

cglobal alignstubbed_pixel_hadamard_ac_16x16_avx
STACKALIGNSTUB pixel_hadamard_ac_16x16_avx,2

cglobal alignstubbed_pixel_hadamard_ac_16x8_avx
STACKALIGNSTUB pixel_hadamard_ac_16x8_avx,2

cglobal alignstubbed_pixel_hadamard_ac_8x16_avx
STACKALIGNSTUB pixel_hadamard_ac_8x16_avx,2

cglobal alignstubbed_pixel_hadamard_ac_8x8_avx
STACKALIGNSTUB pixel_hadamard_ac_8x8_avx,2

cglobal alignstubbed_pixel_hadamard_ac_16x16_xop
STACKALIGNSTUB pixel_hadamard_ac_16x16_xop,2

cglobal alignstubbed_pixel_hadamard_ac_16x8_xop
STACKALIGNSTUB pixel_hadamard_ac_16x8_xop,2

cglobal alignstubbed_pixel_hadamard_ac_8x16_xop
STACKALIGNSTUB pixel_hadamard_ac_8x16_xop,2

cglobal alignstubbed_pixel_hadamard_ac_8x8_xop
STACKALIGNSTUB pixel_hadamard_ac_8x8_xop,2

cglobal alignstubbed_mc_chroma_sse2
STACKALIGNSTUB mc_chroma_sse2,9

cglobal alignstubbed_mc_chroma_sse2_misalign
STACKALIGNSTUB mc_chroma_sse2_misalign,9

cglobal alignstubbed_mc_chroma_ssse3
STACKALIGNSTUB mc_chroma_ssse3,9

cglobal alignstubbed_mc_chroma_ssse3_cache64
STACKALIGNSTUB mc_chroma_ssse3_cache64,9

cglobal alignstubbed_mc_chroma_avx
STACKALIGNSTUB mc_chroma_avx,9

cglobal alignstubbed_deblock_v_luma_avx
STACKALIGNSTUB deblock_v_luma_avx,5

cglobal alignstubbed_deblock_h_luma_avx
STACKALIGNSTUB deblock_h_luma_avx,5

cglobal alignstubbed_deblock_v_chroma_avx
STACKALIGNSTUB deblock_v_chroma_avx,5

cglobal alignstubbed_deblock_v_luma_intra_avx
STACKALIGNSTUB deblock_v_luma_intra_avx,4

cglobal alignstubbed_deblock_h_luma_intra_avx
STACKALIGNSTUB deblock_h_luma_intra_avx,4

cglobal alignstubbed_deblock_v_chroma_intra_avx 
STACKALIGNSTUB deblock_v_chroma_intra_avx,4

cglobal alignstubbed_deblock_h_chroma_intra_avx 
STACKALIGNSTUB deblock_h_chroma_intra_avx,4

%if HIGH_BIT_DEPTH
 cglobal alignstubbed_deblock_h_chroma_mbaff_avx
 STACKALIGNSTUB deblock_h_chroma_mbaff_avx,5

 cglobal alignstubbed_deblock_h_chroma_intra_mbaff_avx
 STACKALIGNSTUB deblock_h_chroma_intra_mbaff_avx,4
%endif

cglobal alignstubbed_deblock_v_luma_sse2
STACKALIGNSTUB deblock_v_luma_sse2,5

cglobal alignstubbed_deblock_h_luma_sse2
STACKALIGNSTUB deblock_h_luma_sse2,5

cglobal alignstubbed_deblock_v_chroma_sse2
STACKALIGNSTUB deblock_v_chroma_sse2,5

cglobal alignstubbed_deblock_v_luma_intra_sse2
STACKALIGNSTUB deblock_v_luma_intra_sse2,4

cglobal alignstubbed_deblock_h_luma_intra_sse2
STACKALIGNSTUB deblock_h_luma_intra_sse2,4

cglobal alignstubbed_deblock_v_chroma_intra_sse2
STACKALIGNSTUB deblock_v_chroma_intra_sse2,4

cglobal alignstubbed_deblock_h_chroma_intra_sse2
STACKALIGNSTUB deblock_h_chroma_intra_sse2,4

cglobal alignstubbed_intra_sad_x9_4x4_ssse3 
STACKALIGNSTUB intra_sad_x9_4x4_ssse3,3

cglobal alignstubbed_intra_satd_x9_4x4_ssse3
STACKALIGNSTUB intra_satd_x9_4x4_ssse3,3

cglobal alignstubbed_intra_sad_x9_8x8_ssse3 
STACKALIGNSTUB intra_sad_x9_8x8_ssse3,5

cglobal alignstubbed_intra_sad_x9_4x4_sse4
STACKALIGNSTUB intra_sad_x9_4x4_sse4,3

cglobal alignstubbed_intra_satd_x9_4x4_sse4
STACKALIGNSTUB intra_satd_x9_4x4_sse4,3

cglobal alignstubbed_intra_sad_x9_8x8_sse4
STACKALIGNSTUB intra_sad_x9_8x8_sse4,5

cglobal alignstubbed_intra_satd_x9_4x4_xop
STACKALIGNSTUB intra_satd_x9_4x4_xop,3

%if HIGH_BIT_DEPTH
 cglobal alignstubbed_intra_sad_x9_4x4_xop
 STACKALIGNSTUB intra_sad_x9_4x4_xop,3
 
 cglobal alignstubbed_intra_sad_x9_8x8_xop
 STACKALIGNSTUB intra_sad_x9_8x8_xop,5
%endif

%if HIGH_BIT_DEPTH
 cglobal alignstubbed_deblock_h_chroma_intra_mbaff_sse2
 STACKALIGNSTUB deblock_h_chroma_intra_mbaff_sse2,4
%endif

%endif